{-# OPTIONS -Wall -Werror -fno-warn-name-shadowing #-}
module Main where

import Data.Char (ord, chr)
import Data.List (intercalate, partition)
import Text.XML.HaXml
import Text.XML.HaXml.Pretty
import Text.XML.HaXml.Types
import Text.XML.HaXml.Html.Pretty (attribute)
import Text.PrettyPrint.HughesPJ (render)

data Mapping = Mapping [Char] [LaTeX]
data LaTeX = LaTeX String | LaTeXMath String deriving Show

main :: IO ()
main = readFile "ent.xml" >>= putStrLn . documentToHaskell . xmlParse ""

documentToHaskell :: Document -> String
documentToHaskell (Document _ _ (Elem "isoent" _ c) _) =
    let (errors, sequences) = partitionEithers (concatMap contentToHaskell c) in
    "module Text.LaTeX.Entities where\n\nimport Data.Char (chr)\n\ndata Mapping = Mapping [Char] [LaTeX]\ndata LaTeX = LaTeX String | LaTeXMath String\n\n" ++
    "latexEntities = [\n  " ++
    intercalate ",\n  " (map sequenceToString sequences) ++ "]\n\n" ++
    case errors of
      [] -> ""
      _ -> "\n\n{-\nErrors:\n" ++ intercalate "\n-}\n" errors
documentToHaskell (Document _ _ (Elem name _ _) _) = error $ "Unexpected element: " ++ show name

partitionEithers :: [Either a b] -> ([a], [b])
partitionEithers xs = 
    foldr f ([], []) xs
    where f (Left x) (lefts, rights) = (x : lefts, rights)
          f (Right x) (lefts, rights) = (lefts, x : rights)

sequenceToString :: Mapping -> String
sequenceToString (Mapping unicode latex) =
    "Mapping [" ++ intercalate ", " (map (\ c -> "chr " ++ show (ord c)) unicode) ++ "] " ++ show latex

contentToHaskell :: Content -> [Either String Mapping]
contentToHaskell (CElem (Elem "char" _ xs)) =
    case collect ["entity", "unicode", "latex", "plain", "comment", "escaped"] xs :: ([[Content]], [Content]) of
      ([_entities, unicodes, latexes, _, _, _], other) -> 
          case filter isContent other of
            [] -> case unicodeValues (filter isContent unicodes) of
                    -- Some entities don't correspond to any unicode character, I guess.
                    Left _error -> [{- Left (error ++ "\n" ++ concatMap (render . content) xs) -}]
                    Right unicode ->
                        map (latexSequence unicode) (filter isContent latexes)
            xs -> error $ "Unexpected elements in char: " ++ concatMap (render . content) xs
      _ -> error "collect"
contentToHaskell x@(CElem (Elem _ _ _)) = error $ "contentToHaskell - unexpected element: " ++ render (content x)
contentToHaskell (CString _ _) = [] -- Left ("CString " ++ show s)
contentToHaskell (CRef _) = [] -- Left "{- CRef -}"
contentToHaskell (CMisc _) = [] -- Left "{- CMisc -}"


isContent :: Content -> Bool
isContent (CString _ _) = False
isContent _ = True

isString :: Content -> Bool
isString (CString _ _) = True
isString _ = False

unicodeValues :: [Content] -> Either String [Char]
unicodeValues xs =
    let ([unicodes], _other) = collect ["unicode"] xs in
    case partitionEithers (map value unicodes) of
      -- (_, []) -> Left "No unicode values"
      ([], values) -> Right values
      (errors, _) -> Left (concat errors)
    where
      value (CElem (Elem "unicode" attrs _)) =
          case lookup "value" attrs of
            Nothing -> Left "No unicode value"
            Just (AttValue xs) -> Right (chr (read ("0x" ++ takeWhile (/= '"') (drop 7 (concatMap (render . attribute) [("value", AttValue xs)])))))
      value _ = error "unicodeValues"

latexSequence :: [Char] -> Content -> Either String Mapping
latexSequence unicode (CElem (Elem "latex" _attrs xs)) =
    case collect ["seq", "mathseq"] xs of
      ([seqs, mathseqs], other) ->
          if all isString other
          then Right (Mapping unicode (map (LaTeX . sequence) seqs ++ map (LaTeXMath . sequence) mathseqs))
          else Left "Unexpected content in latex"
      _ -> error "collect"
    where
      sequence (CElem (Elem _ _attrs sequence)) = concatMap (render . content) sequence
      sequence _ = error "sequence"
latexSequence _ x = error $ "latexSequence - unexpected: " ++ render (content x)

rightsOnly :: [Either a b] -> [b]
rightsOnly (Right x : more) = x : rightsOnly more
rightsOnly (_ : more) = rightsOnly more
rightsOnly [] = []

isRight :: Either a b -> Bool
isRight (Right _) = True
isRight _ = False

collect :: [String] -> [Content] -> ([[Content]], [Content])
collect names content =
    foldr f ([], content) names
    where f :: String -> ([[Content]], [Content]) -> ([[Content]], [Content])
          f name (parts, xs) = let (part, xs') = partition (nameIs name) xs in (part : parts, xs')

nameIs :: Name -> Content -> Bool
nameIs s (CElem (Elem name _ _)) = s == name
nameIs _ _ = False
